@@ -3,7 +3,7 @@
  
   This file is part of mod_python. See COPYRIGHT file for details.
 
-  $Id: apache.py,v 1.13 2000/07/04 17:43:02 gtrubetskoy Exp $
+  $Id: apache.py,v 1.19 2000/08/28 21:33:41 gtrubetskoy Exp $
 
 """
 
@@ -12,6 +12,7 @@ import string
 import traceback
 import time
 import os
+import pdb
 import stat
 import imp
 import types
@@ -61,13 +62,30 @@ class CallBack:
         return obj
 
 
+    class HStack:
+        """
+        The actual stack string lives in the request object so
+        it can be manipulated by both apache.py and mod_python.c
+        """
+        
+        def __init__(self, req):
+            self.req = req
+
+        def pop(self):
+            handlers = string.split(self.req.hstack)
+            if not handlers:
+                return None
+            else:
+                self.req.hstack = string.join(handlers[1:], " ")
+                return handlers[0]
+
     def Dispatch(self, req, htype):
         """
         This is the handler dispatcher.
         """
 
         # be cautious
-        result = HTTP_INTERNAL_SERVER_ERROR
+        result, handler = HTTP_INTERNAL_SERVER_ERROR, ""
 
         # request
         self.req = req
@@ -77,10 +95,14 @@ class CallBack:
         debug = config.has_key("PythonDebug")
 
         try:
-            # cycle through the handlers
-            handlers = string.split(config[htype])
 
-            for handler in handlers:
+            # cycle through the handlers
+            dirs = req.get_all_dirs()
+
+            hstack = self.HStack(req)
+        
+            handler = hstack.pop()
+            while handler:
 
                 # split module::handler
                 l = string.split(handler, '::', 1)
@@ -105,7 +127,7 @@ class CallBack:
                         if sys.path != newpath:
                             sys.path = newpath
                 else:
-                    dir = req.get_dirs()[htype]
+                    dir = req.get_all_dirs()[htype]
                     if dir not in sys.path:
                         sys.path[:0] = [dir]
 
@@ -116,11 +138,18 @@ class CallBack:
                 object = self.resolve_object(module, object_str)
 
                 # call the object
-                result = object(req)
+                if config.has_key("PythonEnablePdb"):
+                    if config["PythonEnablePdb"]:
+                        result = pdb.runcall(object, req)
+                else:
+                    result = object(req)
 
                 # stop cycling through handlers
                 if result != OK:
                     break
+                
+                handler = hstack.pop()
+
 
         except SERVER_RETURN, value:
             # SERVER_RETURN indicates a non-local abort from below
@@ -198,6 +227,7 @@ class CallBack:
         finally:
             # erase the traceback
             etb = None
+            return DONE
 
 def import_module(module_name, req=None):
     """ 
@@ -392,11 +422,21 @@ class CGIStdout(NullIO):
 
         if not self.headers_sent:
             self.headers = self.headers + s
-            ss = string.split(self.headers, '\n\n', 1)
+
+            # are headers over yet?
+            headers_over = 0
+
+            # first try RFC-compliant CRLF
+            ss = string.split(self.headers, '\r\n\r\n', 1)
             if len(ss) < 2:
-                # headers not over yet
-                pass
+                # second try with \n\n
+                ss = string.split(self.headers, '\n\n', 1)
+                if len(ss) >= 2:
+                    headers_over = 1
             else:
+                headers_over = 1
+                    
+            if headers_over:
                 # headers done, process them
                 string.replace(ss[0], '\r\n', '\n')
                 lines = string.split(ss[0], '\n')
@@ -410,7 +450,7 @@ class CGIStdout(NullIO):
                         self.req.content_type = v
                         self.req.headers_out[h] = v
                     else:
-                        self.req.headers_out[h] = v
+                        self.req.headers_out.add(h, v)
                 self.req.send_http_header()
                 self.headers_sent = 1
                 # write the body if any at this point
@@ -553,6 +593,12 @@ OK = REQ_PROCEED = 0
 DONE = -2
 DECLINED = REQ_NOACTION = -1
 
+# constants for get_remote_host
+REMOTE_HOST = 0
+REMOTE_NAME = 1
+REMOTE_NOLOOKUP = 2
+REMOTE_DOUBLE_REV = 3
+
 REQ_ABORTED = HTTP_INTERNAL_SERVER_ERROR
 REQ_EXIT = "REQ_EXIT"         
 SERVER_RETURN = "SERVER_RETURN"
